@@ -1,3 +1,4 @@
+#define XC_WANT_COMPAT_MAP_FOREIGN_API
 #include <xenctrl.h>
 #include <xc_private.h>
 #include <xc_core.h>
@@ -6,9 +7,6 @@
 
 #include "xg_save_restore.h"
 
-#undef ARRAY_SIZE /* We shouldn't be including xc_private.h */
-#define ARRAY_SIZE(a) (sizeof (a) / sizeof ((a)[0]))
-
 static xc_interface *xch;
 
 int help_func(int argc, char *argv[])
