@@ -29,6 +29,7 @@
 #include <string.h>
 #include <unistd.h>
 #include <regex.h>
+#include <xen-tools/libs.h>
 
 #include "xenstat_priv.h"
 
@@ -69,7 +70,7 @@ void getBridge(char *excludeName, char *result, size_t resultLen)
 	struct dirent *de;
 	DIR *d;
 
-	char tmp[256] = { 0 };
+	char tmp[512] = { 0 };
 
 	d = opendir("/sys/class/net");
 	while ((de = readdir(d)) != NULL) {
@@ -78,8 +79,14 @@ void getBridge(char *excludeName, char *result, size_t resultLen)
 				sprintf(tmp, "/sys/class/net/%s/bridge", de->d_name);
 
 				if (access(tmp, F_OK) == 0) {
-					strncpy(result, de->d_name, resultLen - 1);
-					result[resultLen - 1] = 0;
+					/*
+					 * Do not use strncpy to prevent compiler warning with
+					 * gcc >= 10.0
+					 * If de->d_name is longer then resultLen we truncate it
+					 */
+					memset(result, 0, resultLen);
+					memcpy(result, de->d_name, MIN(strnlen(de->d_name,
+									NAME_MAX),resultLen - 1));
 				}
 		}
 	}
@@ -219,8 +226,11 @@ int parseNetDevLine(char *line, char *iface, unsigned long long *rxBytes, unsign
 				}
 				else
 				/* There were errors when parsing this directly in RE. strpbrk() helps */
-				if (iface != NULL)
-					strcpy(iface, strpbrk(tmp, "abcdefghijklmnopqrstvuwxyz0123456789"));
+				if (iface != NULL) {
+					char *tmp2 = strpbrk(tmp, "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789");
+					if (tmp2 != NULL)
+						strcpy(iface, tmp2);
+				}
 
 				memset(tmp, 0, matches[i].rm_eo - matches[i].rm_so);
 			}
@@ -261,7 +271,7 @@ int xenstat_collect_networks(xenstat_node * node)
 {
 	/* Helper variables for parseNetDevLine() function defined above */
 	int i;
-	char line[512] = { 0 }, iface[16] = { 0 }, devBridge[16] = { 0 }, devNoBridge[16] = { 0 };
+	char line[512] = { 0 }, iface[16] = { 0 }, devBridge[16] = { 0 }, devNoBridge[17] = { 0 };
 	unsigned long long rxBytes, rxPackets, rxErrs, rxDrops, txBytes, txPackets, txErrs, txDrops;
 
 	struct priv_data *priv = get_priv_data(node->handle);
@@ -301,7 +311,7 @@ int xenstat_collect_networks(xenstat_node * node)
 
 	/* We get the bridge devices for use with bonding interface to get bonding interface stats */
 	getBridge("vir", devBridge, sizeof(devBridge));
-	snprintf(devNoBridge, 16, "p%s", devBridge);
+	snprintf(devNoBridge, sizeof(devNoBridge), "p%s", devBridge);
 
 	while (fgets(line, 512, priv->procnetdev)) {
 		xenstat_domain *domain;
@@ -433,13 +443,15 @@ int xenstat_collect_vbds(xenstat_node * node)
 		ret = sscanf(dp->d_name, "%3s-%u-%u", buf, &domid, &vbd.dev);
 		if (ret != 3)
 			continue;
+		if (!(strstr(buf, "vbd")) && !(strstr(buf, "tap")))
+			continue;
 
 		if (strcmp(buf,"vbd") == 0)
 			vbd.back_type = 1;
 		else if (strcmp(buf,"tap") == 0)
 			vbd.back_type = 2;
 		else
-			continue;
+			vbd.back_type = 0;
 
 		domain = xenstat_node_domain(node, domid);
 		if (domain == NULL) {
@@ -450,36 +462,29 @@ int xenstat_collect_vbds(xenstat_node * node)
 			continue;
 		}
 
-		if((read_attributes_vbd(dp->d_name, "statistics/oo_req", buf, 256)<=0)
-		   || ((ret = sscanf(buf, "%llu", &vbd.oo_reqs)) != 1))
+		if (vbd.back_type == 1 || vbd.back_type == 2)
 		{
-			continue;
-		}
 
-		if((read_attributes_vbd(dp->d_name, "statistics/rd_req", buf, 256)<=0)
-		   || ((ret = sscanf(buf, "%llu", &vbd.rd_reqs)) != 1))
+			vbd.error = 0;
+
+			if ((read_attributes_vbd(dp->d_name, "statistics/oo_req", buf, 256)<=0) ||
+				((ret = sscanf(buf, "%llu", &vbd.oo_reqs)) != 1) ||
+				(read_attributes_vbd(dp->d_name, "statistics/rd_req", buf, 256)<=0) ||
+				((ret = sscanf(buf, "%llu", &vbd.rd_reqs)) != 1) ||
+				(read_attributes_vbd(dp->d_name, "statistics/wr_req", buf, 256)<=0) ||
+				((ret = sscanf(buf, "%llu", &vbd.wr_reqs)) != 1) ||
+				(read_attributes_vbd(dp->d_name, "statistics/rd_sect", buf, 256)<=0) ||
+				((ret = sscanf(buf, "%llu", &vbd.rd_sects)) != 1) ||
+				(read_attributes_vbd(dp->d_name, "statistics/wr_sect", buf, 256)<=0) ||
+				((ret = sscanf(buf, "%llu", &vbd.wr_sects)) != 1))
+			{
+				vbd.error = 1;
+			}
+		}
+		else
 		{
-			continue;
+			vbd.error = 1;
 		}
-
-		if((read_attributes_vbd(dp->d_name, "statistics/wr_req", buf, 256)<=0)
-		   || ((ret = sscanf(buf, "%llu", &vbd.wr_reqs)) != 1))
-		{
-			continue;
-		}
-
-		if((read_attributes_vbd(dp->d_name, "statistics/rd_sect", buf, 256)<=0)
-		   || ((ret = sscanf(buf, "%llu", &vbd.rd_sects)) != 1))
-		{
-			continue;
-		}
-
-		if((read_attributes_vbd(dp->d_name, "statistics/wr_sect", buf, 256)<=0)
-		   || ((ret = sscanf(buf, "%llu", &vbd.wr_sects)) != 1))
-		{
-			continue;
-		}
-
 		if ((xenstat_save_vbd(domain, &vbd)) == NULL) {
 			perror("Allocation error");
 			return 0;
