@@ -30,7 +30,6 @@
     System Programming Guide; Section 9.11. (1997 edition - PPro).
 */
 
-#include <xen/config.h>
 #include <xen/init.h>
 #include <xen/lib.h>
 #include <xen/smp.h>
@@ -94,9 +93,9 @@ static void __init set_num_var_ranges(void)
 		rdmsrl(MSR_MTRRcap, config);
 	} else if (is_cpu(AMD))
 		config = 2;
-	else if (is_cpu(CYRIX) || is_cpu(CENTAUR))
+	else if (is_cpu(CENTAUR))
 		config = 8;
-	num_var_ranges = config & 0xff;
+	num_var_ranges = MASK_EXTR(config, MTRRcap_VCNT);
 }
 
 static void __init init_table(void)
@@ -332,7 +331,7 @@ int mtrr_add_page(unsigned long base, unsigned long size,
 	if ((type == MTRR_TYPE_WRCOMB) && !have_wrcomb()) {
 		printk(KERN_WARNING
 		       "mtrr: your processor doesn't support write-combining\n");
-		return -ENOSYS;
+		return -EOPNOTSUPP;
 	}
 
 	if (!size) {
