@@ -4,9 +4,9 @@
  * x86-specific shutdown handling.
  */
 
-#include <xen/config.h>
 #include <xen/init.h>
 #include <xen/lib.h>
+#include <xen/param.h>
 #include <xen/sched.h>
 #include <xen/smp.h>
 #include <xen/delay.h>
@@ -26,6 +26,7 @@
 #include <asm/mpspec.h>
 #include <asm/tboot.h>
 #include <asm/apic.h>
+#include <asm/guest.h>
 
 enum reboot_type {
         BOOT_INVALID,
@@ -35,6 +36,7 @@ enum reboot_type {
         BOOT_CF9 = 'p',
         BOOT_CF9_PWR = 'P',
         BOOT_EFI = 'e',
+        BOOT_XEN = 'x',
 };
 
 static int reboot_mode;
@@ -50,10 +52,14 @@ static int reboot_mode;
  * pci    Use the so-called "PCI reset register", CF9
  * Power  Like 'pci' but for a full power-cyle reset
  * efi    Use the EFI reboot (if running under EFI)
+ * xen    Use Xen SCHEDOP hypercall (if running under Xen as a guest)
  */
 static enum reboot_type reboot_type = BOOT_INVALID;
-static void __init set_reboot_type(char *str)
+
+static int __init set_reboot_type(const char *str)
 {
+    int rc = 0;
+
     for ( ; ; )
     {
         switch ( *str )
@@ -73,13 +79,33 @@ static void __init set_reboot_type(char *str)
         case 'P':
         case 'p':
         case 't':
+        case 'x':
             reboot_type = *str;
             break;
+        default:
+            rc = -EINVAL;
+            break;
         }
         if ( (str = strchr(str, ',')) == NULL )
             break;
         str++;
     }
+
+    if ( reboot_type == BOOT_EFI && !efi_enabled(EFI_RS) )
+    {
+        printk("EFI reboot selected, but no EFI runtime services available.\n"
+               "Falling back to default reboot type.\n");
+        reboot_type = BOOT_INVALID;
+    }
+
+    if ( reboot_type == BOOT_XEN && !xen_guest )
+    {
+        printk("Xen reboot selected, but Xen hypervisor not detected\n"
+               "Falling back to default\n");
+        reboot_type = BOOT_INVALID;
+    }
+
+    return rc;
 }
 custom_param("reboot", set_reboot_type);
 
@@ -95,6 +121,10 @@ static inline void kb_wait(void)
 static void noreturn __machine_halt(void *unused)
 {
     local_irq_disable();
+
+    if ( reboot_type == BOOT_XEN )
+        xen_hypercall_shutdown(SHUTDOWN_poweroff);
+
     for ( ; ; )
         halt();
 }
@@ -115,21 +145,32 @@ void machine_halt(void)
 
 static void default_reboot_type(void)
 {
-    if ( reboot_type == BOOT_INVALID )
-        reboot_type = efi_enabled ? BOOT_EFI
-                                  : acpi_disabled ? BOOT_KBD
-                                                  : BOOT_ACPI;
+    if ( reboot_type != BOOT_INVALID )
+        return;
+
+    if ( xen_guest )
+        reboot_type = BOOT_XEN;
+    else if ( efi_enabled(EFI_RS) )
+        reboot_type = BOOT_EFI;
+    else if ( acpi_disabled )
+        reboot_type = BOOT_KBD;
+    else
+        reboot_type = BOOT_ACPI;
 }
 
-static int __init override_reboot(struct dmi_system_id *d)
+static int __init override_reboot(const struct dmi_system_id *d)
 {
     enum reboot_type type = (long)d->driver_data;
 
+    if ( type == BOOT_ACPI && acpi_disabled )
+        type = BOOT_KBD;
+
     if ( reboot_type != type )
     {
         static const char *__initdata msg[] =
         {
             [BOOT_KBD]  = "keyboard controller",
+            [BOOT_ACPI] = "ACPI",
             [BOOT_CF9]  = "PCI",
         };
 
@@ -141,7 +182,7 @@ static int __init override_reboot(struct dmi_system_id *d)
     return 0;
 }
 
-static struct dmi_system_id __initdata reboot_dmi_table[] = {
+static const struct dmi_system_id __initconstrel reboot_dmi_table[] = {
     {    /* Handle problems with rebooting on Dell E520's */
         .callback = override_reboot,
         .driver_data = (void *)(long)BOOT_KBD,
@@ -435,6 +476,15 @@ static struct dmi_system_id __initdata reboot_dmi_table[] = {
             DMI_MATCH(DMI_PRODUCT_NAME, "OptiPlex 390"),
         },
     },
+    {    /* Handle problems with rebooting on Dell OptiPlex 9020. */
+        .callback = override_reboot,
+        .driver_data = (void *)(long)BOOT_ACPI,
+        .ident = "Dell OptiPlex 9020",
+        .matches = {
+            DMI_MATCH(DMI_SYS_VENDOR, "Dell Inc."),
+            DMI_MATCH(DMI_PRODUCT_NAME, "OptiPlex 9020"),
+        },
+    },
     {    /* Handle problems with rebooting on the Latitude E6320. */
         .callback = override_reboot,
         .driver_data = (void *)(long)BOOT_CF9,
@@ -462,6 +512,24 @@ static struct dmi_system_id __initdata reboot_dmi_table[] = {
             DMI_MATCH(DMI_PRODUCT_NAME, "Latitude E6520"),
         },
     },
+    {    /* Handle problems with rebooting on Dell PowerEdge R540. */
+        .callback = override_reboot,
+        .driver_data = (void *)(long)BOOT_ACPI,
+        .ident = "Dell PowerEdge R540",
+        .matches = {
+            DMI_MATCH(DMI_SYS_VENDOR, "Dell Inc."),
+            DMI_MATCH(DMI_PRODUCT_NAME, "PowerEdge R540"),
+        },
+    },
+    {    /* Handle problems with rebooting on Dell PowerEdge R740. */
+        .callback = override_reboot,
+        .driver_data = (void *)(long)BOOT_ACPI,
+        .ident = "Dell PowerEdge R740",
+        .matches = {
+            DMI_MATCH(DMI_SYS_VENDOR, "Dell Inc."),
+            DMI_MATCH(DMI_PRODUCT_NAME, "PowerEdge R740"),
+        },
+    },
     { }
 };
 
@@ -541,6 +609,7 @@ void machine_restart(unsigned int delay_millisecs)
         {
         case BOOT_INVALID:
             ASSERT_UNREACHABLE();
+            /* fall through */
         case BOOT_KBD:
             /* Pulse the keyboard reset line. */
             for ( i = 0; i < 100; i++ )
@@ -590,6 +659,15 @@ void machine_restart(unsigned int delay_millisecs)
             }
             reboot_type = BOOT_ACPI;
             break;
+
+        case BOOT_XEN:
+            /*
+             * When running in PV shim mode guest shutdown calls are
+             * forwarded to L0, hence the only way to get here is if a
+             * shim crash happens.
+             */
+            xen_hypercall_shutdown(pv_shim ? SHUTDOWN_crash : SHUTDOWN_reboot);
+            break;
         }
     }
 }
