@@ -10,7 +10,6 @@
  * omap platform has some specific configurations
  */
 
-#include <xen/config.h>
 #include <xen/console.h>
 #include <xen/serial.h>
 #include <xen/init.h>
@@ -25,6 +24,30 @@
 
 #define REG_SHIFT 2
 
+/* Register offsets */
+#define UART_OMAP_EFR    0x02   /* Enhanced feature register */
+#define UART_OMAP_MDR1   0x08   /* Mode definition register 1 */
+#define UART_OMAP_SCR    0x10   /* Supplementary control register */
+#define UART_OMAP_SSR    0x11   /* Supplementary status register */
+#define UART_OMAP_SYSC   0x15   /* System configuration register */
+#define UART_OMAP_TXFIFO_LVL   0x1A   /* TX FIFO level register */
+
+/* Enhanced feature register */
+#define UART_OMAP_EFR_ECB   0x10   /* Enhanced control bit */
+
+/* Mode definition register 1 */
+#define UART_OMAP_MDR1_16X_MODE   0x00   /* UART 16x mode           */
+#define UART_OMAP_MDR1_DISABLE    0x07   /* Disable (default state) */
+
+/* Supplementary control register bitmasks */
+#define UART_OMAP_SCR_RX_TRIG_GRANU1_MASK   (1 << 7)
+
+/* Supplementary status register bitmasks */
+#define UART_OMAP_SSR_TX_FIFO_FULL_MASK   (1 << 0)
+
+/* System configuration register */
+#define UART_OMAP_SYSC_DEF_CONF   0x0d   /* autoidle mode, wakeup is enabled */
+
 #define omap_read(uart, off)       readl((uart)->regs + (off<<REG_SHIFT))
 #define omap_write(uart, off, val) writel((val), (uart)->regs + (off<<REG_SHIFT))
 
@@ -146,7 +169,7 @@ static void fifo_setup(struct omap_uart *uart)
     /*
      * Load the new FIFO triggers and the new DMA mode bit.
      */
-    omap_write(uart, UART_OMAP_SCR, OMAP_UART_SCR_RX_TRIG_GRANU1_MASK);
+    omap_write(uart, UART_OMAP_SCR, UART_OMAP_SCR_RX_TRIG_GRANU1_MASK);
     /*
      * Restore the UART_OMAP_EFR[4] value.
      */
@@ -196,8 +219,8 @@ static void __init omap_uart_init_preirq(struct serial_port *port)
 
     omap_write(uart, UART_OMAP_MDR1, UART_OMAP_MDR1_16X_MODE);
 
-    /* setup iddle mode */
-    omap_write(uart, UART_SYSC, OMAP_UART_SYSC_DEF_CONF);
+    /* setup idle mode */
+    omap_write(uart, UART_OMAP_SYSC, UART_OMAP_SYSC_DEF_CONF);
 }
 
 static void __init omap_uart_init_postirq(struct serial_port *port)
@@ -233,11 +256,20 @@ static int omap_uart_tx_ready(struct serial_port *port)
 {
     struct omap_uart *uart = port->uart;
     uint32_t reg;
+    uint8_t cnt;
 
     reg = omap_read(uart, UART_IER);
     omap_write(uart, UART_IER, reg | UART_IER_ETHREI);
 
-    return omap_read(uart, UART_LSR) & UART_LSR_THRE ? uart->fifo_size : 0;
+    /* Check for empty space in TX FIFO */
+    if ( omap_read(uart, UART_OMAP_SSR) & UART_OMAP_SSR_TX_FIFO_FULL_MASK )
+        return 0;
+
+    /* Check number of data bytes stored in TX FIFO */
+    cnt = omap_read(uart, UART_OMAP_TXFIFO_LVL);
+    ASSERT( cnt >= 0 && cnt <= uart->fifo_size );
+
+    return (uart->fifo_size - cnt);
 }
 
 static void omap_uart_putc(struct serial_port *port, char c)
