@@ -7,13 +7,22 @@
 #ifndef __XEN_CONFIG_H__
 #define __XEN_CONFIG_H__
 
+#include <xen/kconfig.h>
+
 #ifndef __ASSEMBLY__
 #include <xen/compiler.h>
+
+#if defined(CONFIG_ENFORCE_UNIQUE_SYMBOLS) || defined(__clang__)
+# define EMIT_FILE asm ( "" )
+#else
+# define EMIT_FILE asm ( ".file \"" __FILE__ "\"" )
 #endif
+
+#endif
+
 #include <asm/config.h>
 
 #define EXPORT_SYMBOL(var)
-#define EXPORT_SYMBOL_GPL(var)
 
 /*
  * The following log levels are as follows:
@@ -69,29 +78,25 @@
 #define __force
 #define __bitwise
 
+#define KB(_kb)     (_AC(_kb, ULL) << 10)
 #define MB(_mb)     (_AC(_mb, ULL) << 20)
 #define GB(_gb)     (_AC(_gb, ULL) << 30)
 
+#define IS_ALIGNED(val, align) (((val) & ((align) - 1)) == 0)
+
 #define __STR(...) #__VA_ARGS__
 #define STR(...) __STR(__VA_ARGS__)
 
-#ifndef __ASSEMBLY__
-/* Turn a plain number into a C unsigned long constant. */
-#define __mk_unsigned_long(x) x ## UL
-#define mk_unsigned_long(x) __mk_unsigned_long(x)
-#else /* __ASSEMBLY__ */
-/* In assembly code we cannot use C numeric constant suffixes. */
-#define mk_unsigned_long(x) x
-#endif /* !__ASSEMBLY__ */
+/* allow existing code to work with Kconfig variable */
+#define NR_CPUS CONFIG_NR_CPUS
 
-#define fastcall
-#define __cpuinitdata
-#define __cpuinit
+#ifndef CONFIG_DEBUG
+#define NDEBUG
+#endif
 
-#ifdef FLASK_ENABLE
-#define XSM_MAGIC 0xf97cff8c
-/* Maintain statistics on the access vector cache */
-#define FLASK_AVC_STATS 1
+#ifndef ZERO_BLOCK_PTR
+/* Return value for zero-size allocation, distinguished from NULL. */
+#define ZERO_BLOCK_PTR ((void *)-1L)
 #endif
 
 #endif /* __XEN_CONFIG_H__ */
